/**********************************************************************
 *
 *   yacc++ - Generate object-oriented LR parser
 *   ======
 *
 *   Usage: yacc++ [-dgcvn] input-file [-o output-file]
 *
 *   Options:
 *
 *      -d   Process the yacc grammar with debugging turned on
 *      -v   Be verbose about what is being done
 *      -n   Don't insert #line directives in generated files
 *	-k   Keep temporary intermediate files
 *
 **********************************************************************/

#include <stdio.h>
#include <sys/param.h>
#include <varargs.h>
#include <string.h>
#include <strings.h>
#include "parser.h"
#include "args.h"
#include "token.h"

#define DEFAULT_CPLUS_INCLUDE_DIR "/usr/local/cosc/lib/g++-include"
/*#define CPLUS_CPP "/usr/local/gnu/lib/gcc-lib/sparc-sun-sunos4.1/2.6.3/cpp"*/
/*#define CPLUS_CPP "g++"*/
/*#define CPLUS_CPP_OPTS "-E -dD"*/
#define CPLUS_CPP_OPTS "-E"
#define CPLUS_CPP_IN_SUFFIX ".cppin.c"

#define bool int
#define true 1
#define false 0

char *prog;
char usage[MAXPATHLEN+256];

char *root_name;
char *ypp_file_name;
char *cpp_input_name;
char *cpp_file_name;
char *lex_file_name;
char *yacc_file_name;
char *scanner_file_name;
char *parser_file_name;
char *output_file_name;
char *header_file_name;
char *tmp_debug_file_name;
char *debug_file_name;


int cflag, dflag, nflag, kflag, verbose;

FILE *f;

char *remove_suffix(name) char *name; {
  char *dot;

  name = strdup(name);
  dot = strrchr(name, '.');
  if (dot) {
    *dot = 0;
  }
  return name;
}

char *concat(s1, s2) char *s1, *s2; {
  char *s3 = (char *)malloc(strlen(s1) + strlen(s2) + 1);
  strcpy(s3, s1);
  strcat(s3, s2);
  return s3;
}

main(argc, argv) int argc; char *argv[]; {
  prog = argv[0];
  sprintf(usage, "Usage: %s [-dvnk] input-file [-o output-file]");
  argusage = usage;
  scanargs("dvnk", "o");
  cflag = argcount('c');
  dflag = argcount('d');
  kflag = argcount('k');
  verbose = argcount('v');
  nflag = argcount('n');
  syncing = !nflag;
  if (argcount(0) == 0) {
    fprintf(stderr, "%s: No input file specified\n", prog);
    exit(1);
  }
  ypp_file_name = argstring(0, 0);
  root_name = remove_suffix(ypp_file_name);
  lex_file_name = concat(root_name, "_tmp_scanner.l");
  yacc_file_name = concat(root_name, "_tmp_parser.y");
  scanner_file_name = concat(root_name, "_scanner.yy.c");
  parser_file_name = concat(root_name, "_tmp_parser.tab.c");
  tmp_debug_file_name = concat(root_name, "_tmp_parser.output");
  debug_file_name = concat(root_name, ".output");
  header_file_name = concat(root_name, ".H");
  if (argcount('o'))
    output_file_name = argstring('o', 0);
  else
    output_file_name = concat(root_name, ".C");
#ifdef CPLUS_CPP
  cpp_input_name = concat(ypp_file_name, CPLUS_CPP_IN_SUFFIX);
  cpp_file_name = concat(root_name, ".cpp");
  unlink(cpp_input_name);
  if (link(ypp_file_name, cpp_input_name) < 0) {
    perror(cpp_input_name);
    exit(1);
  }
  shell("%s %s %s | sed -e s/%s// > %s", CPLUS_CPP, CPLUS_CPP_OPTS, 
	cpp_input_name, CPLUS_CPP_IN_SUFFIX, cpp_file_name);
#else
  cpp_file_name = ypp_file_name;
#endif
  yyin = fopen(cpp_file_name, "r");
  if (!yyin) {
    perror(ypp_file_name);
    exit(1);
  }
  header = fopen(header_file_name, "w");
  cdecl1 = fopen("yy.cdecl1.tmp", "w");
  cdecl2 = fopen("yy.cdecl2.tmp", "w");
  ytok = fopen("yy.ytok.tmp", "w");
  yunion = fopen("yy.yunion.tmp", "w");
  ytype = fopen("yy.ytype.tmp", "w");
  yprod = fopen("yy.yprod.tmp", "w");
  ldecl = fopen("yy.ldecl.tmp", "w");
  lrule = fopen("yy.lrule.tmp", "w");
  source_file_name = ypp_file_name;
  line_counter = 1;
  yyparse();
  emit_token_lex_rules();
  declare_yyresult(header);
  put_parsing_function_header(header, true);
  fprintf(header, ";\n");
  fclose(yyin);
  fclose(header);
  fclose(cdecl1);
  fclose(cdecl2);
  fclose(ytok);
  fclose(yunion);
  fclose(ytype);
  fclose(yprod);
  fclose(ldecl);
  fclose(lrule);
  make_lex_file();
  make_yacc_file();
  shell("flex %s; mv lex.yy.c %s", lex_file_name, scanner_file_name);
  shell("bison -v %s %s", 
	dflag ? "-t" : "", 
	yacc_file_name);
  shell("mv %s %s",
	parser_file_name,
	output_file_name);
  shell("mv %s %s",
	tmp_debug_file_name,
	debug_file_name);
  if (!kflag)
    shell("rm -f %s_tmp_* yy.*.tmp", root_name);
  return 0;
}

make_lex_file() {
  f = fopen(lex_file_name, "w");
  put("%{");
  put("#include <iostream.h>");
  put("#include <string.h>");
  put("static void yy_count_lines(char *);");
  put("static int yy_ignore = 0;");
  put("#undef YY_INPUT");
  put("#define YY_INPUT(buf, result, max) yy_input(buf, result, max)");
  put("static void yy_input(char *buf, int &result, int max) {");
  /*put("  fprintf(stderr, \"yy_input: reading up to %d chars\\n\", max);");*/
  put("  char c;");
  put("  result = 0;");
  put("  --max;");
  put("  while (result < max && yy_stream->get(c)) {");
  put("    *buf++ = c; ++result;");
  put("    if (c == '\\n') break;");
  put("  }");
  /*put("  fprintf(stderr, \"yy_input: read %d chars\\n\", result);");*/
  put("}");
  put("%}");
  tack("yy.ldecl.tmp");
  put("%%");
  tack("yy.lrule.tmp");
  put("\\n\t{++yy_line;}");
  put("[ \\t]\t");
  put(".\t{yyerror(\"unrecognised character\");}");
  put("%%");
  put("void yy_count_lines(char *text) {");
  put("\tfor (; *text; ++text)");
  put("\t\tif (*text == '\\n') {++yy_line; yy_col = 1;} else ++yy_col;");
  put("}");
  fclose(f);
}

make_yacc_file() {
  f = fopen(yacc_file_name, "w");
  put("%{");
  put("int yylex();");
  put("static void yyerror(char *);");
  tack("yy.cdecl1.tmp");
  if (parser_name)
    fprintf(f, "#include \"%s\"\n", header_file_name);
  else
    tack(header_file_name);
  tack("yy.cdecl2.tmp");
  put("%}");
  tack("yy.ytok.tmp");
  put("%union {");
  put("\tint i;");
  put("\tfloat f;");
  put("\tchar *s;");
  put("\tdouble d;");
  tack("yy.yunion.tmp");
  put("}");
  tack("yy.ytype.tmp");
  put("%%");
  tack("yy.yprod.tmp");
  put("%%");
  fprintf(f, "#include \"%s\"\n", scanner_file_name);
  make_parsing_function();
  if (!parser_name)
    make_main_program();
  fclose(f);
}

char *parsing_function_name() {
  if (parser_name)
    return parser_name;
  else
    return "parser";
}

declare_yyresult(f) FILE *f; {
  fprintf(f, "struct yyresult {\n");
  fprintf(f, "  char *message;\n");
  fprintf(f, "  char *source;\n");
  fprintf(f, "  int line;\n");
  fprintf(f, "  int col;\n");
  fprintf(f, "  char *token;\n");
  fprintf(f, "};\n");
}

put_parsing_function_header(f, definition) FILE *f; bool definition; {
  fprintf(f, 
	  "%s *%s(class istream &s, yyresult &r, char *source%s)", 
	  start_symbol, parsing_function_name(),
	  definition ? " = 0" : "");
}

make_parsing_function() {
  put("#include <std.h>");
  put("#include <errno.h>");
  put("#include <iostream.h>");
  put("static yyresult *yy_result;");
  put("static void yyerror(char *mess) {");
  put("  yy_result->message = mess;");
  put("  yy_result->source = yy_source;");
  put("  yy_result->line = yy_line;");
  put("  yy_result->col = yy_col;");
  put("  yy_result->token = yytext;");
  put("}");
  put_parsing_function_header(f, false);
  put("{");
  put("  yy_stream = &s;");
  put("  yy_result = &r;");
  put("  yy_result->message = 0;");
  put("  yy_result->source = 0;");
  put("  yy_result->line = 0;");
  put("  yy_result->col = 0;");
  put("  yy_result->token = 0;");
  put("  yy_source = source;");
  put("  yy_line = 1;");
  put("  yy_col = 1;");
  put("  if (yy_current_buffer)");
  put("    yyrestart(0);");
  put("  if (yyparse())");
  put("    return 0;");
  put("  return yytree;");
  put("}");
}

make_main_program() {
  /*
  put("int main(int argc, char *argv[]) {");
  if (dflag) {
    put("  if (argc > 1 && strcmp(argv[1], \"-d\") == 0) {");
    put("    ++argv; --argc;");
    put("    yydebug = 1;");
    put("  }");
  }
  put("  char *fname = 0;");
  put("  if (argc > 1) {");
  put("    fname = argv[1];");
  put("  if (parse(0, fname))");
  put("    yytree->main();");
  put("}\n");
  */
}

put(s) char *s; {
  fprintf(f, "%s\n", s);
}

tack(name) char *name; {
  char buf[1024];
  FILE *g = fopen(name, "r");
  while (fgets(buf, 1024, g))
    fputs(buf, f);
  fclose(g);
}

shell(va_alist) va_dcl {
  va_list ap;
  char *format;
  char buf[4*MAXPATHLEN];
  int status;

  va_start(ap);
  format = va_arg(ap, char *);
  vsprintf(buf, format, ap);
  if (verbose)
    fprintf(stderr, "%s\n", buf);
  status = system(buf) >> 8;
  if (status)
    exit(status);
}

yyerror(mess) char *mess; {
  extern char *yytext;
  fprintf(stderr, "\"%s\", line %d, near '%s': %s\n", 
	  source_file_name, line_counter, yytext, mess);
  exit(1);
}
