%{

#include <stdio.h>
#include <string.h>
#include "parser.h"
#include "token.h"

#define MAX_ITEMS_PER_ALTERNATIVE 1024

char *builtin_type_name[] = {0, "int", "float", "char*", "double"};
char *builtin_union_member[] = {0, "i", "f", "s"};

char *lhs;
int alt_counter;
int item_counter;
int ignore_counter;
char member_flag[MAX_ITEMS_PER_ALTERNATIVE];
FILE *cfile;
char *nodeclass = "yynode";
char *superclass;
char *classname;

%}

%union {
	char *s;
	int i;
}

%token PARSER TOKEN IGNORE NESTED INT FLOAT DOUBLE CHAR NODE CLASS RULE
%token EQUALS SEMI DOLLAR DOTDOT COLON STAR BAR
%token LBRACE RBRACE
%token IDENT PATTERN INTEGER CTOKEN

%type <s> IDENT PATTERN CTOKEN dotdot_pattern
%type <i> INTEGER type nested


%%

grammar	: 	{
		start_symbol = 0;
		fprintf(cdecl1, "#include <iostream.h>\n");
  		fprintf(cdecl1, "static istream *yy_stream;\n");
		fprintf(cdecl1, "static char *yy_source;\n");
		fprintf(cdecl1, "static int yy_line;\n");
		fprintf(cdecl1, "static int yy_col;\n");
		fprintf(header, "struct yynode {\n");
		fprintf(header, "\tchar *yysource;\n");
		fprintf(header, "\tint yyline;\n");
		fprintf(header, "\tyynode();\n");
		fprintf(header, "};\n");
		fprintf(cdecl2, "yynode::yynode() {\n");
		fprintf(cdecl2, "\tyysource = yy_source;\n");
		fprintf(cdecl2, "\tyyline = yy_line;\n");
		fprintf(cdecl2, "}\n");
		}
	  preamble
	  precode 
	  tokendefs
	  nodedef
	  classdefs
	  productions
		{
		if (start_symbol == 0) {
			yyerror("No main production!");
			}
		fprintf(cdecl2, "%s *yytree;\n", start_symbol);
		fprintf(ytype, "%%start yystart\n");
		fprintf(yprod, "yystart : %s {\n", start_symbol);
		fprintf(yprod, "\t\tyytree = $1;\n");
		fprintf(yprod, "\t}\n");
		fprintf(yprod, ";\n");
		}
	  postcode
	;

preamble: PARSER IDENT SEMI {
		parser_name = $2;
		}
	|
	;

precode	: 	{
		cfile = cdecl1;
		lsync(cfile);
		}
	  LBRACE ccode RBRACE
	|
	;

postcode
	: 	{
		cfile = cdecl2;
		lsync(cfile);
		}
	  LBRACE ccode RBRACE
	|
	;

tokendefs
	: tokendefs tokendef
	|
	;

tokendef
	: TOKEN type IDENT EQUALS PATTERN SEMI 
		{
		int kind = $2;
		char *type = builtin_type_name[kind];
		char *name = $3;
		char *pattern = $5;
		char *member = builtin_union_member[kind];
		fprintf(ytok, "%%token %s\n", name);
		if (type) {
			fprintf(ytype, "%%type <%s> %s\n", member, name);
			}
		declare_token(name, kind, type, pattern);
		}
	| IGNORE nested PATTERN dotdot_pattern SEMI
		{
		int nested = $2;
		char *pat1 = $3;
		char *pat2 = $4;
		if (pat2 == 0) {
			fprintf(lrule, "%s\t{yy_count_lines(yytext);}\n", 
				pat1);
			}
		else {
			char state[32];
			sprintf(state, "IGNORE%d", ignore_counter++);
			fprintf(ldecl, "%%x %s\n", state);
			fprintf(lrule, "%s\t{BEGIN(%s); ++yy_ignore;}\n", 
				pat1, state);
			if (nested)
				fprintf(lrule, "<%s>%s\t{++yy_ignore;}\n",
					state, pat1);
			fprintf(lrule, 
			    "<%s>%s\t{if (!--yy_ignore) BEGIN(INITIAL);}\n", 
				state, pat2);
			fprintf(lrule, "<%s>\\n\t{++yy_line;}\n", state);
			fprintf(lrule, "<%s>.\n", state);
			}
		}
	;

nested	: NESTED	{$$ = 1;}
	|		{$$ = 0;}
	;

dotdot_pattern
	: DOTDOT PATTERN
			{$$ = $2;}
	|		{$$ = 0;}
	;

type	: INT		{$$ = intType;}
	| FLOAT		{$$ = floatType;}
	| DOUBLE	{$$ = doubleType;}
	| CHAR STAR	{$$ = stringType;}
	|		{$$ = noType;}
	;

nodedef	: NODE IDENT {
		cfile = header;
		lsync(cfile);
		nodeclass = strdup($2);
		fprintf(cfile, "struct %s : yynode {\n", nodeclass);
		}
	  LBRACE ccode RBRACE {
		lsync(cfile);
	  	fprintf(cfile, "};\n");
		}
	  SEMI
	|
	;

classdefs
	: classdefs classdef
	|
	;

classdef
	: classintro {
		cfile = header;
		lsync(cfile);
		fprintf(cfile, "struct %s : %s {\n", classname, superclass);
	  	}
	  LBRACE ccode RBRACE {
		lsync(cfile);
	  	fprintf(cfile, "};\n");
		}
	  SEMI
	;

classintro
	: IDENT CLASS IDENT {
		superclass = $1;
		classname = $3;
		}
	| CLASS IDENT {
		superclass = nodeclass;
		classname = $2;
		}
	;

productions
	: productions production
	| production
	;

production
	: prodintro {
		if (start_symbol == 0)
			start_symbol = lhs;
		alt_counter = 1;
		fprintf(yunion, "\t%s *u_%s;\n", lhs, lhs);
		fprintf(ytype, "%%type <u_%s> %s\n", lhs, lhs);
		fprintf(yprod, "%s :\n", lhs);
		}
	  mainclass COLON alternatives SEMI 
		{fprintf(yprod, ";\n");}
	;

prodintro
	: RULE IDENT {
		superclass = nodeclass;
		lhs = $2;
		}
	| IDENT RULE IDENT {
		superclass = $1;
		lhs = $3;
		}
	;

mainclass
	: 	{
		cfile = header;
		lsync(cfile);
		fprintf(cfile, "struct %s : %s {\n", lhs, superclass);
		} 
	  maincode
		{
		lsync(cfile);
	  	fprintf(cfile, "};\n");
		}
	;

maincode
	: LBRACE ccode RBRACE
	|
	;

alternatives
	: alternatives BAR {fprintf(yprod, "|\n");} alternative
	| alternative
	;

alternative 
	:	{
		lsync(cdecl2);
		fprintf(cdecl2, "struct %s_%d : %s {\n", 
			lhs, alt_counter, lhs);
		item_counter = 1;
		}
	  items subclass {
		int i;
		fprintf(yprod, "\t\t{\n");
		lsync(yprod);
		fprintf(yprod, 
			"%s_%d *yyn = new %s_%d; ", 
				lhs, alt_counter, lhs, alt_counter);
		for (i = 1; i < item_counter; i++)
			if (member_flag[i]) fprintf(yprod, 
				"yyn->yyval%d = $%d; ", i, i);
		fprintf(yprod, "$$ = yyn; ");
		fprintf(yprod, "} ");
		lsync(cdecl2);
		fprintf(cdecl2, "};\n");
		++alt_counter;
		}
	;

subclass
	: 	{
		cfile = cdecl2;
		lsync(cfile);
		}
	  LBRACE ccode RBRACE
	|
	;

items	: items item
	|
	;

item	: IDENT {
		char *name = $1;
		char *type;
		if (!find_token_by_name(name)) { /* nonterminal */
			type = $1;
			lsync(cdecl2);
			fprintf(cdecl2, "\tstruct %s *yyval%d;\n", 
				type, item_counter);
			member_flag[item_counter] = 1;
			}
		else if (type = token_name_to_token_type(name)) { 
			/* token with a value */
			lsync(cdecl2);
			fprintf(cdecl2, "\t%s yyval%d;\n", type, item_counter);
			member_flag[item_counter] = 1;
			}
		else {
			/* else token with no value */
			member_flag[item_counter] = 0;
			}
		fprintf(yprod, "%s ", name);
		++item_counter;
		}
	| PATTERN {
		char *pat = $1;
		char *name = pattern_to_token_name(pat);
		char *type = token_name_to_token_type(name); 
		fprintf(yprod, "%s ", name);
		if (type) {
			lsync(cdecl2);
			fprintf(cdecl2, "\t%s yyval%d;\n", type, item_counter);
			}
		member_flag[item_counter] = 0;
		++item_counter;
		}
	;

ccode	: ccode citem
	|
	;

citem	: CTOKEN		{fprintf(cfile, "%s", $1);}
	| INTEGER		{fprintf(cfile, "%d", $1);}
	| DOLLAR INTEGER	{fprintf(cfile, "yyval%d", $2);}
	| LBRACE		{fputc('{', cfile);} 
	  ccode 
	  RBRACE		{fputc('}', cfile);}
	;

%%

#include "scanner.yy.c"

lsync(f) FILE *f; {
	/*fprintf(stderr, "Syncing to line %d of %s\n", 
		line_counter, source_file_name);*/
	if (syncing)
		fprintf(f, "\n#line %d \"%s\"\n", 
			line_counter, source_file_name);
}
