letter		[A-Za-z_]
digit		[0-9]
ident		{letter}({letter}|{digit})*
integer		{digit}+
string		\"([^\\\"\n]|\\.)*\"
charconst	\'([^\\\'\n]|\\.)*\'
stdname		"<"[^\>\n]*">"
punct		[^A-Za-z0-9_{}\#\$\" \t\n]+
space		[ \t]+

hash_line	"#"{gap}"line"?{gap}{integer}{gap}({string}|{stdname}).*\n
pp_directive	"#".*
gap		[ \t]*

%x CCode Comment Pattern

%{

#include <sys/param.h>

#define rstring(token)	yylval.s = strdup(yytext); return token;
#define rint(token)	yylval.i = atoi(yytext); return token;

int paren_counter;
int brace_counter;
int old_state;

void process_hashline();

%}

%%

<INITIAL>{hash_line}		{process_hashline();}
<CCode,Comment>{hash_line}	{process_hashline();}

"parser"	{return PARSER;}
"token"		{return TOKEN;}
"ignore"	{return IGNORE;}
"nested"	{return NESTED;}
"int"		{return INT;}
"float"		{return FLOAT;}
"double"	{return DOUBLE;}
"char"		{return CHAR;}
"node"		{return NODE;}
"class"		{return CLASS;}
"="		{return EQUALS;}
";"		{return SEMI;}
"$"		{return DOLLAR;}
":"		{return COLON;}
"*"		{return STAR;}
"|"		{return BAR;}
".."		{return DOTDOT;}

<INITIAL,CCode>"rule"		{return RULE;}

{ident}		{rstring(IDENT);}
{string}	{rstring(PATTERN);}
{integer}	{rint(INTEGER);}

"("		{BEGIN(Pattern);
		 	paren_counter = 1;
		 	yymore();}
<Pattern>"("		{++paren_counter; yymore();}
<Pattern>"\("		{yymore();}
<Pattern>{string}	{yymore();}
<Pattern>[^\(\)]	{yymore();}
<Pattern>")/("		{yymore();}
<Pattern>")"		{if (--paren_counter != 0) {yymore();} else {
		 BEGIN(INITIAL);
		 rstring(PATTERN);}}

"/*"		{old_state = INITIAL;
		 BEGIN(Comment);}
<Comment>[^\*\n]+
<Comment>\n		{count_line();}
<Comment>"*"[^\/]
<Comment>"*/"	{BEGIN(old_state);}

<INITIAL,CCode>"{"			{BEGIN(CCode);
						 ++brace_counter;
						 return LBRACE;}
<CCode>"/*"					{old_state = CCode;
						 BEGIN(Comment);}
<CCode>{ident}|{string}|{charconst}|{punct}|{space}	{rstring(CTOKEN);}
<CCode>{integer}				{rint(INTEGER);}
<CCode>"$"					{return DOLLAR;}
<CCode>{pp_directive}				{rstring(CTOKEN);}
<CCode>\n					{count_line();
						 rstring(CTOKEN);}
<INITIAL,CCode>"}"				{if (--brace_counter == 0)
					 BEGIN(INITIAL);
					 return RBRACE;}

\n		{count_line();}

[ \t]+

<INITIAL,CCode,Comment,Pattern>\"	{yyerror("Unmatched quote");}
<INITIAL,CCode,Comment,Pattern>[.\n]	{yyerror("Unrecognised character");}

%%

count_line() {
	/*fprintf(stderr, "Counting line %d\n", line_counter);*/
	++line_counter;
}

void process_hashline() {
	char name[MAXPATHLEN];
	if (sscanf(yytext, "# line %d \"%[^\"]", &line_counter, name) < 2)
	  sscanf(yytext, "# %d \"%[^\"]", &line_counter, name);
	source_file_name = strdup(name);
}
